@@ -19,11 +19,16 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include "config.h"
+
 #include <stdlib.h>
 
 #include "pangoxft-private.h"
 #include "X11/Xft/XftFreetype.h"
+
+#include "pango-layout.h"
 #include "pango-modules.h"
+#include "pango-utils.h"
 
 #define PANGO_XFT_FONT(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_XFT_FONT, PangoXftFont))
 #define PANGO_XFT_FONT_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), PANGO_TYPE_XFT_FONT, PangoXftFontClass))
@@ -32,11 +37,18 @@
 
 #define PANGO_XFT_UNKNOWN_FLAG 0x10000000
 
-typedef struct _PangoXftFontClass   PangoXftFontClass;
+typedef struct _PangoXftFontClass    PangoXftFontClass;
+typedef struct _PangoXftMetricsInfo  PangoXftMetricsInfo;
 
 struct _PangoXftFontClass
 {
-  PangoFontClass parent_class;
+  PangoFontClass  parent_class;
+};
+
+struct _PangoXftMetricsInfo
+{
+  const char       *sample_str;
+  PangoFontMetrics *metrics;
 };
 
 static PangoFontClass *parent_class;	/* Parent class structure for PangoXftFont */
@@ -91,6 +103,7 @@ pango_xft_font_get_type (void)
 static void 
 pango_xft_font_init (PangoXftFont *xfont)
 {
+  xfont->metrics_by_lang = NULL;
   xfont->in_cache = FALSE;
 }
 
@@ -384,18 +397,70 @@ pango_xft_picture_render (Display          *display,
 }
 
 static PangoFontMetrics *
-pango_xft_font_get_metrics (PangoFont        *font,
-			    PangoLanguage    *language)
+pango_xft_font_get_metrics (PangoFont     *font,
+			    PangoLanguage *language)
 {
-  PangoFontMetrics *metrics = pango_font_metrics_new ();
-  XftFont *xft_font = pango_xft_font_get_font (font);
+  PangoXftFont *xfont = PANGO_XFT_FONT (font);
+  PangoXftMetricsInfo *info = NULL; /* Quiet gcc */
+  GSList *tmp_list;      
 
-  metrics->ascent = PANGO_SCALE * xft_font->ascent;
-  metrics->descent = PANGO_SCALE * xft_font->descent;
-  metrics->approximate_digit_width = PANGO_SCALE * xft_font->max_advance_width;
-  metrics->approximate_char_width = PANGO_SCALE * xft_font->max_advance_width;
+  const char *sample_str = pango_language_get_sample_string (language);
+  
+  tmp_list = xfont->metrics_by_lang;
+  while (tmp_list)
+    {
+      info = tmp_list->data;
+      
+      if (info->sample_str == sample_str)    /* We _don't_ need strcmp */
+	break;
 
-  return metrics;
+      tmp_list = tmp_list->next;
+    }
+
+  if (!tmp_list)
+    {
+      PangoLayout *layout;
+      PangoRectangle extents;
+      PangoContext *context;
+      XftFont *xft_font = pango_xft_font_get_font (font);
+      Display *display;
+  
+      _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
+      context = pango_xft_get_context (display, 0);
+
+      info = g_new (PangoXftMetricsInfo, 1);
+      info->sample_str = sample_str;
+      info->metrics = pango_font_metrics_new ();
+
+      info->metrics->ascent = PANGO_SCALE * xft_font->ascent;
+      info->metrics->descent = PANGO_SCALE * xft_font->descent;
+      info->metrics->approximate_char_width = 
+        info->metrics->approximate_digit_width = 
+        PANGO_SCALE * xft_font->max_advance_width;
+
+      xfont->metrics_by_lang = g_slist_prepend (xfont->metrics_by_lang, 
+                                                info);
+
+      pango_context_set_language (context, language);
+      layout = pango_layout_new (context);
+      pango_layout_set_font_description (layout, xfont->description);
+
+      pango_layout_set_text (layout, sample_str, -1);      
+      pango_layout_get_extents (layout, NULL, &extents);
+      
+      info->metrics->approximate_char_width = 
+        extents.width / g_utf8_strlen (sample_str, -1);
+
+      pango_layout_set_text (layout, "0123456789", -1);      
+      pango_layout_get_extents (layout, NULL, &extents);
+      
+      info->metrics->approximate_digit_width = extents.width / 10;
+
+      g_object_unref (G_OBJECT (layout));
+      g_object_unref (G_OBJECT (context));
+    }
+
+  return pango_font_metrics_ref (info->metrics);
 }
 
 static void
@@ -413,6 +478,13 @@ pango_xft_font_dispose (GObject *object)
   G_OBJECT_CLASS (parent_class)->dispose (object);
 }
 
+static void
+free_metrics_info (PangoXftMetricsInfo *info)
+{
+  pango_font_metrics_unref (info->metrics);
+  g_free (info);
+}
+
 static void
 pango_xft_font_finalize (GObject *object)
 {
@@ -431,6 +503,9 @@ pango_xft_font_finalize (GObject *object)
 
   pango_font_description_free (xfont->description);
 
+  g_slist_foreach (xfont->metrics_by_lang, (GFunc)free_metrics_info, NULL);
+  g_slist_free (xfont->metrics_by_lang);  
+
   if (xfont->xft_font)
     XftFontClose (display, xfont->xft_font);
   else
@@ -456,7 +531,6 @@ pango_xft_font_get_coverage (PangoFont     *font,
   FT_Face face;
   PangoCoverage *coverage;
   Display *display;
-  int i;
   
   _pango_xft_font_map_get_info (xfont->fontmap, &display, NULL);
 
@@ -467,21 +541,38 @@ pango_xft_font_get_coverage (PangoFont     *font,
   if (coverage)
     return pango_coverage_ref (coverage);
 
-  /* Ugh, this is going to be SLOW */
-
   face = pango_xft_font_get_face (font);
 
   coverage = pango_coverage_new ();
-  for (i = 0; i < G_MAXUSHORT; i++)
-    {
-      FT_UInt glyph = FT_Get_Char_Index (face, i);
-
-      if (glyph && glyph < face->num_glyphs)
-	pango_coverage_set (coverage, i, PANGO_COVERAGE_EXACT);
-    }
+#ifdef HAVE_FT_GET_FIRST_CHAR
+  {
+    FT_UInt gindex;
+    FT_ULong charcode;
+    
+    charcode = FT_Get_First_Char (face, &gindex);
+    while (gindex)
+      {
+	pango_coverage_set (coverage, charcode, PANGO_COVERAGE_EXACT);
+	charcode = FT_Get_Next_Char (face, charcode, &gindex);
+      }
+  }
+#else  
+  /* Ugh, this is going to be SLOW */
+  {
+    gunichar wc;
+    
+    for (wc = 0; wc < G_MAXUSHORT; wc++)
+      {
+	FT_UInt glyph = FT_Get_Char_Index (face, wc);
+	
+	if (glyph && glyph < face->num_glyphs)
+	  pango_coverage_set (coverage, wc, PANGO_COVERAGE_EXACT);
+      }
+  }
+#endif
 
   _pango_xft_font_map_set_coverage (xfont->fontmap, filename, coverage);
-      
+ 
   return coverage;
 }
 
@@ -586,6 +677,52 @@ pango_xft_font_find_shaper (PangoFont     *font,
   return (PangoEngineShape *)pango_map_get_engine (shape_map, ch);
 }
 
+static gboolean
+set_unicode_charmap (FT_Face face)
+{
+  int charmap;
+
+  for (charmap = 0; charmap < face->num_charmaps; charmap++)
+    if (face->charmaps[charmap]->encoding == ft_encoding_unicode)
+      {
+	FT_Error error = FT_Set_Charmap(face, face->charmaps[charmap]);
+	return error == FT_Err_Ok;
+      }
+
+  return FALSE;
+}
+
+static void
+load_fallback_font (PangoXftFont *xfont)
+{
+  Display *display;
+  int screen;
+  XftFont *xft_font;
+  
+  _pango_xft_font_map_get_info (xfont->fontmap, &display, &screen);
+      
+  xft_font = XftFontOpen (display,  screen,
+                          XFT_FAMILY, XftTypeString, "sans",
+                          XFT_ENCODING, XftTypeString, "glyphs-fontspecific",
+                          XFT_CORE, XftTypeBool, False,
+                          XFT_SIZE, XftTypeDouble, (double)pango_font_description_get_size (xfont->description)/PANGO_SCALE,
+                          NULL);
+  
+  if (!xft_font)
+    {
+      g_warning ("Cannot open fallback font, nothing to do");
+      exit (1);
+    }
+
+  if (!set_unicode_charmap (xft_font->u.ft.font->face))
+    {
+      g_warning ("Cannot set unicode character map for fallback font, nothing to do");
+      exit (1);
+    }
+
+  xfont->xft_font = xft_font;
+}
+
 /**
  * pango_xft_font_get_font:
  * @font: a #PangoFont.
@@ -600,9 +737,6 @@ pango_xft_font_get_font (PangoFont *font)
   PangoXftFont *xfont;
   Display *display;
   int screen;
-  FT_Face face;
-  FT_Error error;
-  int charmap;
 
   g_return_val_if_fail (PANGO_XFT_IS_FONT (font), NULL);
   
@@ -624,31 +758,26 @@ pango_xft_font_get_font (PangoFont *font)
 	  g_warning ("Cannot open font file for font %s", name);
 	  g_free (name);
 	  
-	  xfont->xft_font = XftFontOpen (display,  screen,
-					 XFT_FAMILY, XftTypeString, "sans",
-					 XFT_ENCODING, XftTypeString, "glyphs-fontspecific",
-					 XFT_CORE, XftTypeBool, False,
-					 XFT_SIZE, XftTypeDouble, (double)pango_font_description_get_size (xfont->description)/PANGO_SCALE,
-					 NULL);
+	  load_fallback_font (xfont);
 	}
-      if (!xfont->xft_font)
+      else
 	{
-	  g_warning ("Cannot open fallback font, nothing to do");
-	  exit (1);
+	  /* There should be a unicode encoding, since we queried for it */
+	  if (!set_unicode_charmap (xfont->xft_font->u.ft.font->face))
+	    {
+	      gchar *name = pango_font_description_to_string (xfont->description);
+	      
+	      g_warning ("Cannot load unicode character map for font %s", name);
+	      g_free (name);
+	      
+	      XftFontClose (display, xfont->xft_font);
+	      xfont->xft_font = NULL;
+		
+	      load_fallback_font (xfont);
+	    }
 	}
-
-      face = xfont->xft_font->u.ft.font->face;
-
-      /* There should be a unicode encoding, since we queried for it */
-      for (charmap = 0; charmap < face->num_charmaps; charmap++)
-	if (face->charmaps[charmap]->encoding == ft_encoding_unicode)
-	  break;
-  
-      g_assert (charmap != face->num_charmaps);
-  
-      error = FT_Set_Charmap(face, face->charmaps[charmap]);
     }
-  
+
   return xfont->xft_font;
 }
 
